"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisCache = exports.notFoundPlaceholder = void 0;
const ioredis_1 = require("ioredis");
const singleflight_1 = require("@zcong/singleflight");
const codec_1 = require("./codec");
const utils_1 = require("./utils");
const stat_1 = require("./stat");
const keyStringer_1 = require("./keyStringer");
exports.notFoundPlaceholder = '*';
let promClient = undefined;
let requestsCounter;
let hitCounter;
let errorsCounter;
let hitNotFoundCacheCounter;
const defaultIsNotFound = (val) => val === null;
const defaultOption = {
    codec: 'json',
    notFoundExpire: 10,
    isNOtFound: defaultIsNotFound,
    name: 'default',
};
class RedisCache {
    constructor(option) {
        this.option = option;
        this.onError = utils_1.noopHandler;
        this.sf = new singleflight_1.Singleflight();
        this.stat = (0, stat_1.createStat)();
        this.option = {
            ...defaultOption,
            ...option,
        };
        if (option.withPrometheus && !promClient) {
            promClient = (0, utils_1.loadPackage)('prom-client', 'withPrometheus', () => require('prom-client'));
            this.setupPrometheus();
        }
    }
    async cacheFn(key, fn, expire, codec, thisArg) {
        this.incrRequestsCounter(1);
        try {
            const [val, isNOtFound] = await this.get(key, codec);
            // if is not found cache, return null
            if (isNOtFound) {
                this.incrHitCounter(1);
                this.incrHitNotFoundCacheCounter(1);
                return null;
            }
            if (val !== null) {
                this.incrHitCounter(1);
                return val;
            }
        }
        catch (err) {
            this.incrErrorsCounter(1);
            this.onError({
                key,
                error: err,
                action: 'get cache',
            });
        }
        const [data, fresh] = await this.sf.doWithFresh(key, async () => {
            const res = await (0, utils_1.bindThis)(fn, thisArg)();
            try {
                await this.set(key, res, expire, codec);
            }
            catch (err) {
                this.incrErrorsCounter(1);
                this.onError({
                    key,
                    error: err,
                    action: 'set cache',
                });
            }
            return res;
        });
        if (!fresh) {
            this.incrHitCounter(1);
        }
        return data;
    }
    cacheWrapper(keyPrefix, fn, expire, codec, keyStringer = keyStringer_1.combineKeyStringer, thisArg) {
        return ((...args) => {
            const cacheKey = RedisCache.joinKey(keyPrefix, keyStringer(...args));
            return this.cacheFn(cacheKey, () => (0, utils_1.bindThis)(fn, thisArg)(...args), expire, codec);
        });
    }
    async deleteFnCache(keyPrefix, args, keyStringer = keyStringer_1.combineKeyStringer) {
        const cacheKey = RedisCache.joinKey(keyPrefix, keyStringer(...args));
        await this.delete(cacheKey);
    }
    /**
     * get data from cache
     * @param key
     * @param codec
     * @returns [val, isNotFound]
     */
    async get(key, codec) {
        const val = await this.option.redis.get(this.buildKey(key));
        if (val === exports.notFoundPlaceholder) {
            return [null, true];
        }
        if (!val) {
            return [null, false];
        }
        try {
            const vv = this.getCodecByName(codec).decode(val);
            return [vv, false];
        }
        catch (err) {
            // delete invalid cache
            await this.delete(key);
            this.incrErrorsCounter(1);
            this.onError({
                key,
                error: err,
                action: 'decode cache',
            });
            return [null, false];
        }
    }
    async set(key, val, expire, codec) {
        if (this.option.isNOtFound(val)) {
            await this.option.redis.set(this.buildKey(key), exports.notFoundPlaceholder, 'EX', this.option.notFoundExpire);
            return;
        }
        const vv = this.getCodecByName(codec).encode(val);
        await this.option.redis.set(this.buildKey(key), vv, 'EX', expire);
    }
    async delete(...keys) {
        if (keys.length === 0) {
            return;
        }
        await Promise.all(keys.map((key) => this.option.redis.del(this.buildKey(key))));
    }
    async clean(match = '*', count = 100) {
        const normalizedPatterns = this.buildKey(match);
        if (this.option.redis instanceof ioredis_1.Cluster) {
            // get only master nodes to scan for deletion,
            // if we get slave nodes, it would be failed for deletion.
            const nodes = this.option.redis.nodes('master');
            await Promise.all(nodes.map((node) => (0, utils_1.redisScanDel)(node, normalizedPatterns, count)));
        }
        else {
            await (0, utils_1.redisScanDel)(this.option.redis, normalizedPatterns, count);
        }
    }
    getStat() {
        return {
            ...this.stat,
        };
    }
    resetStat() {
        this.stat = (0, stat_1.createStat)();
    }
    buildKey(key) {
        return RedisCache.joinKey(this.option.prefix, key);
    }
    getCodecByName(codecName) {
        return (0, codec_1.getCodec)(codecName ?? this.option.codec);
    }
    setupPrometheus() {
        if (!this.option.withPrometheus) {
            return;
        }
        requestsCounter = new promClient.Counter({
            name: 'node_cache_requests_total',
            help: 'Total number of requests to the cache.',
            labelNames: ['name'],
        });
        hitCounter = new promClient.Counter({
            name: 'node_cache_hits_total',
            help: 'Total number of requests to the cache that were a hit.',
            labelNames: ['name'],
        });
        errorsCounter = new promClient.Counter({
            name: 'node_cache_errors_total',
            help: 'Total number of errors to the cache.',
            labelNames: ['name'],
        });
        hitNotFoundCacheCounter = new promClient.Counter({
            name: 'node_cache_hit_not_found_cache_total',
            help: 'Total number of requests to the cache that hit a not found cache.',
            labelNames: ['name'],
        });
    }
    incrRequestsCounter(val) {
        this.stat.requestsCounter++;
        this.incrCounter(requestsCounter, val);
    }
    incrHitCounter(val) {
        this.stat.hitCounter++;
        this.incrCounter(hitCounter, val);
    }
    incrErrorsCounter(val) {
        this.stat.errorsCounter++;
        this.incrCounter(errorsCounter, val);
    }
    incrHitNotFoundCacheCounter(val) {
        this.stat.hitNotFoundCacheCounter++;
        this.incrCounter(hitNotFoundCacheCounter, val);
    }
    incrCounter(counter, val) {
        if (!counter) {
            return;
        }
        counter.inc({ name: this.option.name }, val);
    }
    static joinKey(...keys) {
        return keys.filter(Boolean).join(':');
    }
}
exports.RedisCache = RedisCache;
//# sourceMappingURL=cache.js.map